#
# Copyright (c) 2000 Going Virtual, LLC. All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Contact Information:
#
# Going Virtual, LLC
# 10800 Independence Pointe Pkwy
# Matthews, NC 28105
# www.goingv.com
#
# Robert E. Tierney 
# bob.tierney@goingv.com 
# 1 (704) 849-0731 x109
#
#
__doc__ = """Going Virtual Interbase Database Python API

Python Database API Specification 2.0 Compliant

This is the one and only module the the user must include to use this API.
Others may be imported as needed for catching exceptions, testing column
types, and so on.  The user should not include the gvibBase module. This
is the C-level interface that implements the ConnectionType and CursorType
objects that are used by this Python API to interface with the Interbase
API library.

$Id: gvib.py,v 1.8 2000/09/15 18:28:35 btierney Exp $
"""
__version__ = '$Revision: 1.8 $'[11:-2]

import gvibBase
import gvibExceptions
import gvibMessages
import time
import types

from gvibTypes import *
from gvibCursor import Cursor
from gvibMessages import messageTable

apilevel     = "2.0"
threadsafety = 1
paramstyle   = "qmark"
parammarker  = "?"

#
# Some Default Connection Parameters
#
# _default_cols => The default number of input and output XSQLVAR structures 
#                  to allocate when a Cursor object is created. Setting to
#                  a value that will eliminate or at least minimize the
#                  number of times the XSQLVAR structures must be reallocated
#                  whenever differing SQL statements are executed.
#
# _default_tpb  => The default transaction parameters used after a connection
#                  is made to the database and a transaction is started.
#                  IMPORTANT:  isc_tpb_version3 MUST be first.
#
# _default_dpb  => This mapping object allows the user to pass a small subset
#                  of database parameters to the code that attaches to the
#                  database.  Implemented mostly to allow the user to specify
#                  a character set or message file other than the defaults.
#                  The key values must be one of the values in gvibTypes.py
#                  and the parameters must be passed as a keyword = value
#                  syntax if transaction parameters are not given.
#                  For example:  
#
#                  tr_parms = whatever
#                  db_parms = { gvibTypes.isc_dpb_lc_ctype : 'DOS437' }
#
#                  cn = gvib.connect(db, user, pwd, tr_parms, db_parms)  or
#                  cn = gvib.connect(db, user, pwd, dpb = db_parms)
#
#                  The supported database parameters are:
#
#                    isc_dpb_num_buffers - Number of cache buffers
#                    isc_dpb_force_write - Write synchrously or asnychronously
#                    isc_dpb_no_reserve  - Reserve extra space on each page
#                    isc_dpb_lc_messages - Message file
#                    isc_dpb_lc_ctype    - Character set
#
#
_default_cols = 20

_default_tpb = (
    isc_tpb_version3,
    isc_tpb_write,
    isc_tpb_concurrency,
    isc_tpb_shared,
    isc_tpb_wait,
    isc_tpb_read_committed,
    isc_tpb_rec_version
)

_default_dpb = {}

def connect(dsn, user, password, tpb = _default_tpb, dpb = _default_dpb):
    """
    connect - Connect to the database and return a connection object.

    This module level function makes a connection to the database and
    then instaniates a Connection object and returns it to the caller.
    An explicit transaction is made after connecting to the database.
    """
    
    if type(tpb) != types.TupleType:
        emsg = messageTable['InvalidTPBType']
        raise gvibExceptions.ProgrammError(ibmsg = emsg)
                          
    if type(dpb) != types.DictionaryType:
        emsg = messageTable['InvalidDPBType']
        raise gvibExceptions.Error(ibmsg = emsg)

    ib_connect = gvibBase.IBConnect(dsn, user, password, tpb, dpb)

    py_connect = Connection(ib_connect)

    return py_connect


class Date:
    """
    Date - A lightweight date object for use when dealing with DATE columns.
  
    When a date column appears in a select clause the actual data will be
    returned as one of these objects.  When a date column appears in a 
    predicate clause the user must instantiate one of these objects with
    the appropriate date value.
    """

    daysPerMonth = [(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31), 
                    (31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)]

    def __init__(self, yy = None, mm = 1, dd = 1):
        if yy == None:
	    T = time
            self.year, self.month, self.day = T.localtime(T.time())[:3]
	    return

        from types import *
	validTypes = (IntType, LongType)

	v1 = type(yy) in validTypes
	v2 = type(mm) in validTypes
	v3 = type(dd) in validTypes

        if (v1 + v2 + v3) < 3:
	    emsg = messageTable['invalidDateTypes']
	    raise gvibExceptions.Error(ibmsg = emsg)

        self.year  = yy
	self.month = mm
	self.day   = dd

	if not self.isValidDate():
	    emsg = messageTable['invalidDate']
	    emsg = emsg + ' (%d, %d, %d)' % (yy, mm, dd)
	    raise gvibExceptions.Error(ibmsg = emsg)

    def __str__(self):
        return '%4.4d/%2.2d/%2.2d' % (self.year, self.month, self.day)

    def tuple(self):
        return (self.year, self.month, self.day, 0, 0, 0)

    def isValidDate(self):
        if self.year < 0:
            return 0
    
        if self.month not in range(1, 13):
            return 0;
	
        if self.day not in range(1, 32):
            return 0;
	
	yy = self.year
        isLeapYear = ((yy % 4 == 0) and (yy % 100 != 0 or yy % 400 == 0))
    
        return self.day <= Date.daysPerMonth[isLeapYear][self.month - 1]



class Time:
    """
    Time - A lightweight time object for use when dealing with TIME columns.
  
    When a time column appears in a select clause the actual data will be
    returned as one of these objects.  When a time column appears in a 
    predicate clause the user must instantiate one of these objects with
    the appropriate time value.
    """

    def __init__(self, hh = None, mi = 0, ss = 0):
        if hh == None:
	    T = time
            self.hour, self.minute, self.second = T.localtime(T.time())[3:6] 
	    return

        from types import IntType, LongType
	validTypes = (IntType, LongType)

        v1 = type(hh) in validTypes
        v2 = type(mi) in validTypes
        v3 = type(ss) in validTypes

        if (v1 + v2 + v3) < 3:
	    emsg = messageTable['invalidTimeTypes']
	    raise gvibExceptions.Error(ibmsg = emsg)

        self.hour   = hh
        self.minute = mi
        self.second = ss

	if not self.isValidTime():
	    emsg = messageTable['invalidTime']
	    emsg = emsg + ' (%d %d %d)' %  (hh, mi, ss)
	    raise gvibExceptions.Error(ibmsg = emsg)

    def __str__(self):
        return '%2.2d:%2.2d:%2.2d' % (self.hour, self.minute, self.second)

    def tuple(self):
        return (1, 1, 1, self.hour, self.minute, self.second)

    def isValidTime(self):
        if self.hour not in range(0, 24):
            return 0

        if self.minute not in range(0, 60):
            return 0

        if self.second not in range(0, 60):
            return 0

        return 1


class Timestamp:
    """
    Timestamp - A timestamp object for use when dealing with TIMESTAMP columns.
  
    When a timestamp column appears in a select clause the actual data will be
    returned as one of these objects.  When a timestamp column appears in a 
    predicate clause the user must instantiate one of these objects with
    the appropriate timestamp value.

    """

    def __init__(self, yy = None, mm = 1, dd = 1, hh = 0, mi = 0, ss = 0.0):
        if yy == None:
	    T = time
            ticks = T.time()
            (self.year, self.month, self.day, 
	     self.hour, self.minute, self.second) = T.localtime(ticks)[:6]

            self.second = self.second + (ticks % int(ticks))
	    return

	from types import *
	validTypes = (IntType, LongType)

        v1 = type(yy) in validTypes
        v2 = type(mm) in validTypes
        v3 = type(dd) in validTypes
        v4 = type(hh) in validTypes
        v5 = type(mi) in validTypes
        v6 = type(ss) in validTypes

	if not v6:
	    v6 = type(ss) == FloatType

        if (v1 + v2 + v3 + v4 + v5 + v6) < 6:
	    emsg = messageTable['invalidTimestampTypes']
	    raise gvibExceptions.Error(ibmsg = emsg)

        self.year   = yy
	self.month  = mm
	self.day    = dd
	self.hour   = hh
	self.minute = mi
	self.second = float(ss)

        if not self.isValidTimestamp():
	    emsg1 = messageTable['invalidTimestamp']
	    emsg2 = ' (%d, %d, %d, %d, %d, %f)' % (yy, mm, dd, hh, mi, ss)
	    raise gvibExceptions.Error(ibmsg = emsg1 + emsg2)
    
    def __str__(self):
        return '%4.4d/%2.2d/%2.2d %2.2d:%2.2d:%4.4f' % (self.year, self.month, 
	                         self.day, self.hour, self.minute, self.second)

    def tuple(self):
        return (self.year, self.month, self.day, 
	        self.hour, self.minute, self.second)

    def isValidTimestamp(self):
        if self.year < 0:
            return 0
    
        if self.month not in range(1, 13):
            return 0;
	
        if self.day not in range(1, 32):
            return 0;
	
	yy = self.year
        isLeapYear = ((yy % 4 == 0) and (yy % 100 != 0 or yy % 400 == 0))
    
        if not self.day <= Date.daysPerMonth[isLeapYear][self.month - 1]:
	    return 0

        if self.hour not in range(0, 24):
            return 0

        if self.minute not in range(0, 60):
            return 0

        if int(self.second) not in range(0, 60):
            return 0

        return 1


def DateFromTicks(ticks):
    """
    DateFromTicks - Create a Date object from ticks (from epoch) value.
    """

    return apply(Date, time.localtime(ticks)[:3])


def TimeFromTicks(ticks):
    """
    TimeFromTicks - Create a Time object from ticks (from epoch) value.
    """

    return apply(Time, time.localtime(ticks)[3:6])


def TimestampFromTicks(ticks):
    """
    TimestampFromTicks - Create Timestamp object from ticks (from epoch) value.
    """

    return apply(Timestamp, time.localtime(ticks)[:6])


def binary(string):
    """
    binary - Create a BufferType object from a string value.
    """

    return buffer(string)


def _gvibDateTime(type, yy, mm, dd, hh, mi, ss):
    """
    _gvibDateTime - 
    """

    result = None

    if type == SQL_TYPE_DATE:
        result = Date(yy, mm, dd)
    elif type == SQL_TYPE_TIME:
        result = Time(hh, mi, ss)
    elif type == SQL_TIMESTAMP:
        result = Timestamp(yy, mm, dd, hh, mi, ss)
    else:
        pass
    
    return result


class DBAPITypeObject:
    """
    DBAPITypeObject Class - Used to implement the DB column type objects.
    """

    def __init__(self, *values):
        self.values = values

    def __cmp__(self, other):
        if other in self.values:
	    return 0
        
	if other < self.values:
	    return 1

        return -1
    

ROWID    = DBAPITypeObject(SQL_ROWID)
STRING   = DBAPITypeObject(SQL_TEXT, 
                           SQL_VARYING)
BINARY   = DBAPITypeObject(SQL_BLOB, 
                           SQL_ARRAY, 
			   SQL_QUAD)
NUMBER   = DBAPITypeObject(SQL_SHORT, 
                           SQL_LONG, 
			   SQL_INT64,
			   SQL_FLOAT, 
			   SQL_DOUBLE,
                           SQL_D_FLOAT, 
			   SQL_NUMERIC,
			   SQL_DECIMAL)
DATETIME = DBAPITypeObject(SQL_TYPE_DATE, 
                           SQL_TYPE_TIME, 
			   SQL_TIMESTAMP)


class Connection:
    """
    Connection Class - This class represents a database connection object.

    This class represents a database Connection object.  An instance of this
    class is returned when the user invokes gvib.connect(...). The methods
    of this class allow the user to commit a transaction, rollback a trans-
    action and close a database connection.   This class is also used to
    instantiate a Cursor object.  This is done when the user invokes the
    cursor method against a Connection object.
    """

    __CLOSED_STATE = 0
    __OPEN_STATE   = 1

    def __init__(self, ib_connect):
        self.__connect = ib_connect
	self.__status  = Connection.__OPEN_STATE

    def close(self):
        """
	close - Rollback and close a database connection.

	This method closes an open database connection.  First a rollback is
	is issued against the database and then the database is detached.
	Any uncommited changes will be lost so make sure you do an explict
	commit before closing the connection.
	"""

        self._checkStatus()

        connect = self.__connect

	if connect.tr_handle:
	    rc = gvibBase.isc_rollback_transaction(
	                  connect.status,
	                  connect.tr_handle)

            if rc: gvibBase.IBSetException(connect.status)

	if connect.db_handle:
	    rc = gvibBase.isc_detach_database(
	                  connect.status,
	                  connect.db_handle)

            if rc: gvibBase.IBSetException(connect.status)

        self.__status  = Connection.__CLOSED_STATE

	del self.__connect


    def commit(self):
        """
	commit - Commit any pending changes and retain the transaction.

	This method issues a commit against the database to save any pending
	changes.  The transaction is retained so that the user can continue
	working and commit any number of times, like while doing positioned
	updates.
	"""

        self._checkStatus()

        connect = self.__connect

        rc = gvibBase.isc_commit_retaining(connect.status, connect.tr_handle);

        if rc: gvibBase.IBSetException(connect.status)


    def rollback(self):
        """
	rollback - Rollback a transaction but retain the transaction context.

	This method issues a rollback against the database to discard any
	pending changes.  The transaction is retained so that the user can
	continue working and possibly commit or rollback again at a later time.
	"""

        self._checkStatus()

        connect = self.__connect

        rc = gvibBase.isc_rollback_retaining(
                      connect.status,
                      connect.tr_handle)

        if rc: gvibBase.IBSetException(connect.status)


    def cursor(self, name = None, cols = 0):
        """
	cursor - Return an instance of a Cursor class object.

        This method instantiates a Cursor object and returns it to the user.
	The optional name can be used to associate a name with a cursor.  This
	would then be used to do a positioned update or delete.  The optional
	cols parameter can be used to override the default number of XSQLVAR
	structures that are created when the Cursor object is instantiated.
	"""

        self._checkStatus()

        if cols < 1:
	    global _default_cols
	    cols = _default_cols

        py_cursor = Cursor(self, self.__connect, name, cols)

        return py_cursor


    def _checkStatus(self):
	"""
	_checkStatus - Ensure the connection hasn't been closed.
	"""

        if self.__status == Connection.__CLOSED_STATE:
	    msg = messageTable['badConnect']
	    raise gvibExceptions.Error(ibmsg = msg)


    def dump(self):
        """
        dumpDBInfo - Retrieve and print some information about the database
                     connection.
        """

        result = ' ' * 512

        dbp = chr(isc_info_db_id)               + \
              chr(isc_info_version)             + \
              chr(isc_info_db_sql_dialect)      + \
              chr(isc_info_page_size)           + \
              chr(isc_info_db_size_in_pages)    + \
              chr(isc_info_num_buffers)         + \
              chr(isc_info_reads)               + \
              chr(isc_info_writes)              + \
              chr(isc_info_ods_version)         + \
              chr(isc_info_ods_minor_version)   + \
              chr(isc_info_forced_writes)       + \
              chr(isc_info_logfile)             + \
              chr(isc_info_cur_logfile_name)    + \
              chr(isc_info_sweep_interval)      + \
              chr(isc_info_end)

        connect = self.__connect

        rc = gvibBase.isc_database_info(connect.status, 
                                        connect.db_handle,
                                        len(dbp),
                                        dbp,
                                        len(result),
                                        result)

        if rc: gvibBase.IBSetException(connect.status)

        index = 0
        while ord(result[index]) != isc_info_end:
            curItem = ord(result[index])
            index   = index + 1
            length  = gvibBase.isc_vax_integer(result[index:], 2)
            index   = index + 2
            
            if curItem == isc_info_num_buffers:
                value = gvibBase.isc_vax_integer(result[index:], length)
                print 'Number of Buffers: %d' % value

            elif curItem == isc_info_reads:
                value = gvibBase.isc_vax_integer(result[index:], length)
                print 'Read Count       : %d' % value

            elif curItem == isc_info_writes:
                value = gvibBase.isc_vax_integer(result[index:], length)
                print 'Write Count      : %d' % value

            elif curItem == isc_info_sweep_interval:
                value = gvibBase.isc_vax_integer(result[index:], length)
                print 'Sweep Interval   : %d' % value

            elif curItem == isc_info_forced_writes:
                value = gvibBase.isc_vax_integer(result[index:], length)
                print 'Forced Writes    : %d' % value

            elif curItem == isc_info_page_size:
                value = gvibBase.isc_vax_integer(result[index:], length)
                print 'Page Size        : %d' % value

            elif curItem == isc_info_ods_version:
                value = gvibBase.isc_vax_integer(result[index:], length)
                print 'OD Version Number: %d' % value

            elif curItem == isc_info_ods_minor_version:
                value = gvibBase.isc_vax_integer(result[index:], length)
                print 'OD Minor Version : %d' % value

            elif curItem == isc_info_db_sql_dialect:
                value = gvibBase.isc_vax_integer(result[index:], length)
                print 'SQL Dialect      : %d' % value

            elif curItem == isc_info_db_size_in_pages:
                value = gvibBase.isc_vax_integer(result[index:], length)
                print 'DB Size in Pages : %d' % value

            elif curItem == isc_info_logfile:
                value = gvibBase.isc_vax_integer(result[index:], length)
                print 'Use Log File     : %d' % value

            elif curItem == isc_info_cur_logfile_name:
                print 'Log File Name    : %s' % result[index : index + length]

            elif curItem == isc_info_db_id:
                strlen = ord(result[index + 1])
                lower = index + 2
                upper = lower + strlen
                print 'Database ID      : %s' % result[lower : upper]

            elif curItem == isc_info_version:
                strlen = ord(result[index + 1])
                lower = index + 2
                upper = lower + strlen
                print 'Version          : %s' % result[lower : upper]

            else:
                pass

            index = index + length
        

gvibBase.IBSetDateTimeCallback((_gvibDateTime,));
